import os
import re
import shutil
import uuid
from glob import glob

from PIL import Image


def check_if_valid_image(file: str, return_bool: bool = False):
    try:
        with Image.open(file) as img:
            img.verify()
        return True
    except Exception:
        # couldn't open and verify -> not a valid image
        if return_bool:
            return False
        else:
            raise ValueError(rf"'{file}' does not appear to be a valid path to a PNG file")


def get_file_size(file_path: str) -> int:
    return 0 if not os.path.isfile(file_path) else os.stat(file_path).st_size


def copy_file(from_file: str, to_file: str):
    # skip copy if equals
    if from_file == to_file:
        return
    # create directory if not already exists
    output_dir = os.path.dirname(os.path.abspath(to_file))
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir, exist_ok=True)
    shutil.copy(from_file, to_file)


def compare_and_use_better_option(file_path_1: str, file_path_2: str, destination_file: str) -> None:
    size_option_1 = get_file_size(file_path_1)
    size_option_2 = get_file_size(file_path_2)

    # compression worked -> copy file to final destination
    if check_if_valid_image(file_path_2) and size_option_1 > size_option_2 or \
            not check_if_valid_image(file_path_1):
        copy_file(file_path_2, destination_file)
    # error in output file -> copy source file to destination
    else:
        copy_file(file_path_1, destination_file)


def get_files_in_folder(folder_path: str, ending: str = "*"):
    return glob(os.path.join(folder_path, ending), include_hidden=True)


def get_filename(full_path_to_file: str) -> str:
    filename_with_ending = os.path.basename(full_path_to_file)
    return re.split(r"\.[^.]*$", filename_with_ending)[0]


def print_stats(orig: int, result: int) -> None:
    if orig < 0:
        raise ValueError("orig must be greater than or equal to 0")
    if result < 0:
        raise ValueError("result can't be less than 0")
    orig_size = str(round(orig / 1000000, 2))
    result_size = str(round(result / 1000000, 2))
    percentage = 0 if orig == 0 else str(-1 * round(100 - (result / orig * 100), 2))
    print(f"Compressed File from {orig_size}mb to {result_size}mb ({percentage}%)\n")



import asyncio
import os
import shutil
import sys
import tempfile
from enum import Enum
from io import TextIOWrapper

from .batch_processor import batch_process_files_async
from .config import TESSERACT_PATH
from .converter.image_to_pdf_converter import ImagesToPdfConverter
from .converter.pdf_merger import merge_pdf_files
from .converter.pdf_to_image_converter import PdfToImageConverter
from .cpdfsqueeze_compressor import CPdfSqueezeCompressor
from .file_operations import get_file_size, get_files_in_folder, print_stats
from .png_compressor import PNGCompressor
from .processor import processor


class PDFCompressor:
    Mode = Enum("Mode", ["AUTO", "LOSSLESS", "FORCE_OCR", "NO_OCR"])

    def __init__(
            self,
            compression_mode: int = 5,
            force_ocr: bool = False,
            no_ocr: bool = False,
            tesseract_language: str = "eng",
            default_pdf_dpi: int = 400,
            event_handlers=None
    ):
        self.event_handlers = event_handlers
        if event_handlers is None:
            self.event_handlers = []

        self.__tessdata_prefix = None
        self.__force_ocr = force_ocr

        # configure compressors
        try:
            # lossless compressor
            self.__cpdf_squeeze_compressor = CPdfSqueezeCompressor()
        except ValueError as error:
            self.__cpdf_squeeze_compressor = None
            print(str(error) + " -> skipping compression with cpdfsqueeze.")

        if (not no_ocr and force_ocr) and not os.path.exists(TESSERACT_PATH):
            # if tesseract is definitely necessary
            raise ValueError(
                "TESSERACT_PATH not found."
                "If force-ocr is active tesseract needs to be configured correctly."
            )
        # init lossy compressor
        self.__png_crunch_compressor = PNGCompressor(compression_mode)
        self.__image_to_pdf_converter = ImagesToPdfConverter(
            self.__force_ocr,
            no_ocr,
            tesseract_language,
        )
        self.__pdf_to_image_converter = PdfToImageConverter(
            "png", default_pdf_dpi)

    @processor
    def process_file(self, source_file: str, destination_path: str) -> None:
        if os.path.isdir(destination_path):
            destination_path = os.path.join(destination_path, os.path.basename(source_file))

        start_file_size = get_file_size(source_file)
        temp_folder_01 = tempfile.mkdtemp()
        temp_folder_02 = tempfile.mkdtemp()

        # 1. split pdf into images that can be compressed using crunch
        self.__pdf_to_image_converter.process_file(source_file, temp_folder_01)

        # 2. compress all images in temp_folder
        asyncio.run(
            batch_process_files_async(get_files_in_folder(
                temp_folder_01), temp_folder_02, self.__png_crunch_compressor)
        )

        # 3. convert pngs to pdfs (as separate pages) and optionally apply OCR
        asyncio.run(
            batch_process_files_async(get_files_in_folder(temp_folder_02), temp_folder_02,
                                      self.__image_to_pdf_converter)
        )

        # 4. merge pdf pages into pdf
        if not os.path.exists(os.path.dirname(destination_path)):
            os.makedirs(os.path.dirname(destination_path), exist_ok=True)
        merge_pdf_files(get_files_in_folder(temp_folder_02, "*.pdf"), destination_path)

        # clean up temp folders
        shutil.rmtree(temp_folder_01, ignore_errors=True)
        shutil.rmtree(temp_folder_02, ignore_errors=True)

        # supress normal console outputs
        stdin_buffer = sys.stdin
        sys.stdin = TextIOWrapper(open(os.devnull, "w"))

        if self.__cpdf_squeeze_compressor is not None:
            # compression with cpdf
            self.__cpdf_squeeze_compressor.process_file(destination_path, destination_path)

        if not self.__force_ocr and get_file_size(source_file) < get_file_size(destination_path):
            if self.__cpdf_squeeze_compressor is not None:
                self.__cpdf_squeeze_compressor.process_file(
                    source_file, destination_path)
            if get_file_size(source_file) < get_file_size(destination_path):
                print("File couldn't be compressed.", file=sys.stderr)
                # copy source_file to destination_file
                if not source_file == destination_path:
                    shutil.copy(source_file, destination_path)
            else:
                print(
                    "File couldn't be compressed using crunch cpdf combi. "
                    "However cpdf could compress it. -> No OCR was Created.", file=sys.stderr
                )
        print_stats(start_file_size, get_file_size(destination_path))
        # load normal stdin from buffer
        sys.stdin = stdin_buffer



import os
import subprocess
import sys
from subprocess import CalledProcessError

from .config import ADVPNG_PATH
from .file_operations import copy_file, check_if_valid_image, get_file_size
from .processor import processor


class ADVPNGCompressor:
    def __init__(
            self,
            shrink_rate: int = 2,
            iterations: int = 1,
            extra_args: str = "",
            event_handlers=None
    ):
        """
        Compresses PNG images using advpng.
        :param shrink_rate (from 'https://www.advancemame.it/doc-advpng')
            -0, --shrink-store
                Disable the compression. The file is only stored and not compressed.
                The file is always rewritten also if it's bigger.
            -1, --shrink-fast
                Set the compression level to "fast" using the zlib compressor.
            -2, --shrink-normal
                Set the compression level to "normal" using the libdeflate compressor.
                This is the default level of compression.
            -3, --shrink-extra
                Set the compression level to "extra" using the 7z compressor.
                You can define the compressor iterations with the -i, --iter option.
            -4, --shrink-insane
                Set the compression level to "insane" using the zopfli compressor.
                You can define the compressor iterations with the -i, --iter option.
        :param iterations - amount of repetitions > 0
            more -> better compression but much slower
        :param extra_args: extra args to pass the executable
        :param event_handlers: (list of EventHandler) A list of event handlers. (pre and postprocessors)
        """
        self.event_handlers = event_handlers
        if self.event_handlers is None:
            self.event_handlers = []

        advpng_path = ADVPNG_PATH
        if not os.path.isfile(advpng_path):
            linux_error = "Install it with 'sudo apt install advancecomp'." if os.name != "nt" else ""
            raise FileNotFoundError(rf"advpng path not found at '{advpng_path}'. {linux_error}")

        if shrink_rate < 0 or shrink_rate > 4:
            raise ValueError("shrink_rate needs to be a value in range 0-4")
        if iterations < 0:
            raise ValueError("iterations need to be greater than 0")

        system_extra = "powershell.exe" if os.name == 'nt' else ""
        # compress, shrink-normal, 3 rounds of compression
        advpng_options = " ".join(("--recompress", f"-{shrink_rate}", f"-i {iterations}", extra_args))
        self.advpng_command = f"{system_extra}  {advpng_path} {advpng_options}"

    @processor
    def process_file(self, source_file: str, destination_path: str) -> None:
        if os.path.isdir(destination_path):
            destination_path = os.path.join(destination_path, os.path.basename(source_file))
        copy_file(source_file, destination_path)
        check_if_valid_image(source_file)

        try:
            subprocess.check_output(f"{self.advpng_command} '{destination_path}'",
                                    stderr=subprocess.STDOUT, shell=True)
        except CalledProcessError as cpe:
            print(repr(cpe), file=sys.stderr)
            print("processing failed at the advpng stage. (IGNORE)\n", file=sys.stderr)
        except Exception as e:
            print(repr(e), file=sys.stderr)  # explicitly dont raise e

        if (check_if_valid_image(destination_path, True)
                or get_file_size(source_file) < get_file_size(destination_path)):
            copy_file(source_file, destination_path)



from .cpdfsqueeze_compressor import CPdfSqueezeCompressor
from .advpng_compressor import ADVPNGCompressor
from .pdf_compressor import PDFCompressor
from .png_compressor import PNGCompressor
from .pngquant_compressor import PNGQuantCompressor
from .pngcrush_compressor import PNGCrushCompressor
from .pipeline_processor import PipelineProcessor
from .converter.image_to_pdf_converter import ImagesToPdfConverter
from .converter.pdf_to_image_converter import PdfToImageConverter
from .converter.pdf_merger import merge_pdf_files
from .batch_processor import batch_process_files_async, batch_process_files

__all__ = ["CPdfSqueezeCompressor", "ADVPNGCompressor", "PDFCompressor", "PNGCompressor", "PNGQuantCompressor",
           "PNGCrushCompressor", "PipelineProcessor", "PdfToImageConverter", "ImagesToPdfConverter",
           "batch_process_files", "batch_process_files_async", "merge_pdf_files"]



import os
import sys

from .pipeline_processor import PipelineProcessor
from .CompressionPostprocessor import CompressionPostprocessor
from .advpng_compressor import ADVPNGCompressor
from .file_operations import print_stats, get_file_size
from .pngcrush_compressor import PNGCrushCompressor
from .pngquant_compressor import PNGQuantCompressor
from .processor import processor


class PNGCompressor:
    def __init__(
            self,
            compression_mode: int = 3,
            event_handlers=None
    ):
        """
         Compresses PNG images using a combination of compression tools.
         :param compression_mode: Speed of compression,
            where 0 is the slowest (best quality) and 5 is the fastest.
         :param event_handlers: A list of event handlers. (pre and postprocessors)
         :returns: None
         """

        self.event_handlers = event_handlers
        if self.event_handlers is None:
            self.event_handlers = []

        if compression_mode <= 0 or compression_mode >= 6:
            raise ValueError("Compression mode must be in range 1-5")

        advcomp_iterations = None
        advcomp_shrink_rate = None
        pngquant_max_quality = None
        pngquant_min_quality = None
        pngquant_speed = None
        match compression_mode:
            case 1:
                advcomp_iterations = 3
                advcomp_shrink_rate = 4
                pngquant_max_quality = 80
                pngquant_min_quality = 0
                pngquant_speed = 1
            case 2:
                advcomp_iterations = 2
                advcomp_shrink_rate = 3
                pngquant_max_quality = 85
                pngquant_min_quality = 25
                pngquant_speed = 2
            case 3:
                advcomp_iterations = 2
                advcomp_shrink_rate = 2
                pngquant_max_quality = 85
                pngquant_min_quality = 25
                pngquant_speed = 2
            case 4:
                advcomp_iterations = 1
                advcomp_shrink_rate = 2
                pngquant_max_quality = 90
                pngquant_min_quality = 25
                pngquant_speed = 8
            case 5:
                advcomp_iterations = 1
                advcomp_shrink_rate = 1
                pngquant_max_quality = 99
                pngquant_min_quality = 25
                pngquant_speed = 9

        try:
            self.__advcomp = ADVPNGCompressor(
                shrink_rate=advcomp_shrink_rate,
                iterations=advcomp_iterations,
                event_handlers=[CompressionPostprocessor("advcomp")]
            )
        except FileNotFoundError:
            print("Error: Program advcomp not found, skipped compression with advcomp.", file=sys.stderr)
            self.__advcomp = None

        try:
            self.__pngquant = PNGQuantCompressor(
                speed=pngquant_speed,
                min_quality=pngquant_min_quality,
                max_quality=pngquant_max_quality,
                event_handlers=[CompressionPostprocessor("pngquant")]
            )
        except FileNotFoundError:
            print("Error: Program pngquant not found, skipped compression with pngquant.", file=sys.stderr)
            self.__pngquant = None

        try:
            self.__pngcrush = PNGCrushCompressor(
                event_handlers=[CompressionPostprocessor("pngcrush")]
            )
        except FileNotFoundError:
            print("Error: Program pngcrush not found, skipped compression with pngcrush.", file=sys.stderr)
            self.__pngcrush = None

    @processor
    def process_file(self, source_file: str, destination_path: str) -> None:
        if os.path.isdir(destination_path):
            destination_path = os.path.join(destination_path, os.path.basename(source_file))
        if not os.path.exists(os.path.dirname(destination_path)):
            os.makedirs(os.path.dirname(destination_path), exist_ok=True)

        start_file_size = get_file_size(source_file)
        processor = PipelineProcessor(True, [self.__pngquant, self.__advcomp, self.__pngcrush])
        processor.process_file(source_file, destination_path)
        print_stats(start_file_size, get_file_size(destination_path))



import asyncio
from typing import List


async def batch_process_files_async(file_list: List[str], destination: str, processor):
    loop = asyncio.get_running_loop()
    tasks = [
        loop.run_in_executor(
            None,
            processor.process_file, file, destination)
        for file in file_list
    ]
    return await asyncio.gather(*tasks)


def batch_process_files(file_list: List[str], destination: str, processor):
    for file in file_list:
        processor.process_file(file, destination)



import os

COMPRESSOR_LIB_PATH = os.path.abspath(os.path.join(
    os.path.dirname(__file__), "compressor_lib"
))

# ---------------------------------------
if "WINE_PATH" in os.environ:
    WINE_PATH = os.environ["WINE_PATH"]
else:
    WINE_PATH = os.path.join("/", "usr", "bin", "wine")

# ---------------------------------------
if "ADVPNG_PATH" in os.environ:
    ADVPNG_PATH = os.environ["ADVPNG_PATH"]
elif os.name == "nt":  # Windows default
    ADVPNG_PATH = os.path.join(COMPRESSOR_LIB_PATH, "advpng", "advpng.exe")
else:  # Linux default
    ADVPNG_PATH = os.path.join("/", "usr", "bin", "advpng")

# ---------------------------------------
if "PNGQUANT_PATH" in os.environ:
    PNGQUANT_PATH = os.environ["PNGQUANT_PATH"]
elif os.name == "nt":  # Windows default
    PNGQUANT_PATH = os.path.join(COMPRESSOR_LIB_PATH, "pngquant", "pngquant.exe")
else:  # Linux default
    PNGQUANT_PATH = os.path.join("/", "usr", "bin", "pngquant")

# ---------------------------------------
if "PNGCRUSH_PATH" in os.environ:
    PNGCRUSH_PATH = os.environ["PNGCRUSH_PATH"]
elif os.name == "nt":  # Windows default
    PNGCRUSH_PATH = os.path.join(COMPRESSOR_LIB_PATH, "pngcrush", "pngcrush.exe")
else:  # Linux default
    PNGCRUSH_PATH = os.path.join("/", "usr", "bin", "pngcrush")

# ---------------------------------------
if "CPDFQUEEZE_PATH" in os.environ:
    CPDFQUEEZE_PATH = os.environ["CPDFQUEEZE_PATH"]
else:
    CPDFQUEEZE_PATH = os.path.join(COMPRESSOR_LIB_PATH, "cpdfsqueeze", "cpdfsqueeze.exe")

# ---------------------------------------
if "TESSERACT_PATH" in os.environ:
    TESSERACT_PATH = os.environ["TESSERACT_PATH"]
elif os.name == "nt":  # Windows default
    TESSERACT_PATH = os.path.join(os.path.abspath(os.path.expanduser('~')), "AppData", "Local", "Programs",
                                 "Tesseract-OCR", "tesseract.exe")
else:  # Linux default
    TESSERACT_PATH = os.path.join("/", "usr", "bin", "tesseract")

# ---------------------------------------
if "TESSDATA_PREFIX" in os.environ:
    TESSDATA_PREFIX = os.environ["TESSDATA_PREFIX"]
elif os.name == "nt":  # Windows default
    TESSDATA_PREFIX = "--tessdata-dir '" + os.path.join(os.path.expanduser('~'), 'AppData', 'Local', 'Programs',
                                                        'Tesseract-OCR', 'tessdata')
else:  # Linux default
    TESSDATA_PREFIX = ""



import os
import subprocess

from .config import PNGCRUSH_PATH
from .file_operations import compare_and_use_better_option, check_if_valid_image
from .processor import processor


class PNGCrushCompressor:
    def __init__(
            self,
            event_handlers=None
    ):
        """
        Compresses PNG images using pngcrush.
        :param event_handlers: (list of EventHandler) A list of event handlers. (pre and postprocessors)
        :returns: None
        """

        pngcrush_path = PNGCRUSH_PATH
        self.event_handlers = event_handlers
        if self.event_handlers is None:
            self.event_handlers = []

        if not os.path.isfile(pngcrush_path):
            linux_error = "Install it with 'sudo apt install pngcrush'." if os.name != "nt" else ""
            raise FileNotFoundError(
                rf"pngcrush path not found at '{pngcrush_path}'. {linux_error}")

        system_extra = "powershell.exe" if os.name == 'nt' else ""
        pngcrush_options = "-rem alla -rem text -reduce"  # -brute"
        # TODO add option brute when compression mode is high
        self.pngcrush_command = f"{system_extra} {pngcrush_path} {pngcrush_options}"

    @processor
    def process_file(self, source_file: str, destination_path: str) -> None:
        check_if_valid_image(source_file)
        if os.path.isdir(destination_path):
            destination_path = os.path.join(destination_path, os.path.basename(source_file))
        if not os.path.exists(os.path.dirname(destination_path)):
            os.makedirs(os.path.dirname(destination_path), exist_ok=True)

        subprocess.check_output(rf'{self.pngcrush_command} "{source_file}" "{source_file[:-4] + "-comp.png"}"',
                                stderr=subprocess.STDOUT, shell=True)
        result_file = source_file[:-4] + '-comp.png'
        compare_and_use_better_option(
            source_file, result_file, destination_path)
        if os.path.exists(result_file):
            os.remove(result_file)



import os
import subprocess

from .config import WINE_PATH, CPDFQUEEZE_PATH
from .processor import processor


class CPdfSqueezeCompressor:
    def __init__(
            self,
            use_wine_on_linux: bool = True,
            extra_args: str = "",
            event_handlers=None
    ):
        """
        Pdf Compression via cpdf
        :param use_wine_on_linux:
             True means it runs the command via wine.
                Only should be active when using linux or unix.
                If it is active on Windows, it will be automatically disabled.
            False means it directly runs the executable file at the path.
        :param extra_args: extra args to pass the executable (e.g. -upw <password> for a user password)
        :param event_handlers: (list of EventHandler) A list of event handlers. (pre and postprocessors)
        """
        self.extra_args = extra_args
        self.event_handlers = event_handlers
        if self.event_handlers is None:
            self.event_handlers = []

        self.__cpdfsqueeze_path = CPDFQUEEZE_PATH

        if not os.path.exists(self.__cpdfsqueeze_path):
            raise ValueError(
                rf"cpdfsqueeze_path couldn't be found. '{self.__cpdfsqueeze_path}'")

        # optionally add wine to the command on Linux
        if os.name != "nt" and use_wine_on_linux:
            if not os.path.exists(WINE_PATH):
                raise ValueError(
                    rf"'{WINE_PATH}' wine path couldn't be found. Install it with 'sudo apt install wine'.")
            self.__cpdfsqueeze_path = f"{WINE_PATH} {self.__cpdfsqueeze_path}"

    @processor
    def process_file(self, source_file: str, destination_path: str) -> None:
        if os.path.isdir(destination_path):
            destination_path = os.path.join(destination_path, os.path.basename(source_file))
        # TODO maybe switch to mime type
        if not os.path.exists(source_file) or not source_file.endswith(".pdf"):
            raise ValueError("Only pdf files are accepted")
        if not os.path.exists(os.path.dirname(destination_path)):
            os.makedirs(os.path.dirname(destination_path), exist_ok=True)

        command = f'{self.__cpdfsqueeze_path} "{source_file}" "{destination_path}" {self.extra_args}'
        subprocess.check_output(command, stderr=subprocess.STDOUT, shell=True)



import argparse

from . import CPdfSqueezeCompressor, PDFCompressor, PNGCompressor, PNGCrushCompressor, \
    PNGQuantCompressor, ADVPNGCompressor


def main():
    parser = argparse.ArgumentParser(description="Process a file through a series of processors.")
    parser.add_argument('input_file', type=str, help="The input file to be processed.")
    parser.add_argument('output_file', type=str, help="The output file after processing.")

    # Customize options for each processor
    processor_help = {
        "CPdfSqueezeCompressor": {
            "description": "Lossless PDF Compression.",
            "class": CPdfSqueezeCompressor
        },
        "PDFCompressor": {
            "description": "Lossy PDF Compression, great for scans.",
            "class": PDFCompressor
        },
        "PNGCompressor": {
            "description": "Combined Lossy PNG compression.",
            "class": PNGCompressor
        },
        "ADVPNGCompressor": {
            "description": "Lossy PNG compression.",
            "class": ADVPNGCompressor
        },
        "PNGQuantCompressor": {
            "description": "Lossy PNG compression.",
            "class": PNGQuantCompressor
        },
        "PNGCrushCompressor": {
            "description": "Lossy PNG compression.",
            "class": PNGCrushCompressor
        },
    }

    processor_group = parser.add_mutually_exclusive_group(required=True)

    for i, processor in enumerate(processor_help.keys()):
        processor_group.add_argument(
            f"-{i}",
            f'--{processor.lower()}',
            action='store_true',
            help=processor_help[processor]["description"]
        )

    args = parser.parse_args()

    for processor in processor_help:
        if getattr(args, processor.lower()):
            processor_help[processor]["class"]().process_file(args.input_file, args.output_file)


if __name__ == "__main__":
    main()



import os

# package name PyMuPdf
import fitz

from ..file_operations import get_filename
from ..processor import processor


class PdfToImageConverter:
    SUPPORTED_FILETYPES = ["png", "pnm", "pgm", "pbm", "ppm", "pam", "psd", "ps"]  # TODO test all possible types

    def __init__(
            self,
            file_type_to: str,
            dpi: int = 400,
            event_handlers=None
    ):
        """
        :param file_type_to: resulting image file type
        :param dpi: pdf dpi that should be used when converting to image pixels
        :param event_handlers: A list of event handlers. (pre and postprocessors)
        """
        self.event_handlers = event_handlers
        if self.event_handlers is None:
            self.event_handlers = []

        if file_type_to == "":
            raise ValueError("file_type_to cant be empty!")
        self._file_type_to = self.without_dot_at_the_beginning(file_type_to).lower()
        if self._file_type_to not in self.SUPPORTED_FILETYPES:
            raise ValueError(f"{file_type_to} is not supported.")

        if dpi < 0:
            raise ValueError("default dpi needs to be greater than 0")
        self.__dpi = dpi

    @staticmethod
    def without_dot_at_the_beginning(string):
        return string if not string.endswith(".") else string[1:]

    @processor
    def process_file(self, source_file: str, destination_path: str) -> None:
        # create destination directory if not already exists
        # TODO create proper preprocessor/postprocessors for console output
        print("--splitting pdf into images--")

        # open pdf and split it into rgb-pixel maps -> png
        doc = fitz.open(source_file)
        chars_needed_for_highest_page_number = len(str(len(doc)))

        def get_page_number_string(page_num: int) -> str:
            raw_num = str(page_num)
            return "0" * (chars_needed_for_highest_page_number - len(raw_num)) + raw_num

        for page in doc:
            print(f"** - Finished Page {page.number + 1}/{len(doc)}")
            pix = page.get_pixmap(dpi=self.__dpi)
            page_number = get_page_number_string(page.number + 1)
            pix.save(os.path.join(destination_path, '%s_page_%s.%s' %
                                  (get_filename(source_file), page_number, self._file_type_to)))



import os
from unittest import TestCase

from src.file_crusher.batch_processor import batch_process_files
from src.file_crusher.cpdfsqueeze_compressor import CPdfSqueezeCompressor
from src.file_crusher.pdf_compressor import PDFCompressor


class TestProcessor(TestCase):
    def test_cpdfsqueeze_processor(self):
        destination_file = "tmp.pdf"
        self.assertFalse(os.path.exists(destination_file))
        batch_process_files(["../../test_data/testFile.pdf"], destination_file, CPdfSqueezeCompressor())
        self.assertTrue(os.path.exists(destination_file))
        os.remove(destination_file)

    def test_pdf_crunch_processor(self):
        destination_file = "tmp.pdf"
        self.assertFalse(os.path.exists(destination_file))
        batch_process_files(["../../test_data/testFile.pdf"], destination_file, PDFCompressor())
        self.assertTrue(os.path.exists(destination_file))
        os.remove(destination_file)

    def test_temp_larger_file(self):
        PDFCompressor().process_file(os.path.abspath("../../../Projects/pythonProject/input_file.pdf"), "output.pdf")



import os
import shutil
import tempfile
from unittest import TestCase

from file_crusher import PipelineProcessor


class ProcessorA:
    def process_file(self, source, destination):
        with open(source, 'r') as src, open(destination, 'w') as dst:
            content = src.read()
            dst.write(content + "\nProcessed by A")


class ProcessorB:
    def process_file(self, source, destination):
        with open(source, 'r') as src, open(destination, 'w') as dst:
            content = src.read()
            dst.write(content + "\nProcessed by B")


class ProcessorC:
    def process_file(self, source, destination):
        with open(source, 'r') as src, open(destination, 'w') as dst:
            content = src.read()
            dst.write(content + "\nProcessed by C")


class TestPipelineProcessor(TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.input_file = os.path.join(self.test_dir, 'input.txt')
        self.output_file = os.path.join(self.test_dir, 'output.txt')

        with open(self.input_file, 'w') as f:
            f.write("Initial content")

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_pipeline_all_processors_succeed(self):
        processors = [ProcessorA(), ProcessorB(), ProcessorC()]
        pipeline = PipelineProcessor(True, processors=processors)
        pipeline.process_file(self.input_file, self.output_file)

        with open(self.output_file, 'r') as f:
            content = f.read()

        expected_content = "Initial content\nProcessed by A\nProcessed by B\nProcessed by C"
        self.assertEqual(content.strip(), expected_content)

    def test_pipeline_with_failed_processor(self):
        class FailingProcessor:
            def process_file(self, source, destination):
                raise Exception("Simulated failure")

        processors = [ProcessorA(), FailingProcessor(), ProcessorC()]
        pipeline = PipelineProcessor(True, processors=processors)
        pipeline.process_file(self.input_file, self.output_file)

        with open(self.output_file, 'r') as f:
            content = f.read()

        expected_content = "Initial content\nProcessed by A\nProcessed by C"
        self.assertEqual(content.strip(), expected_content)

    def test_pipeline_with_failed_processor_and_no_skipping(self):
        class FailingProcessor:
            def process_file(self, source, destination):
                raise Exception("Simulated failure")

        processors = [ProcessorA(), FailingProcessor(), ProcessorC()]
        pipeline = PipelineProcessor(False, processors=processors)
        self.assertRaises(Exception, pipeline.process_file, self.input_file, self.output_file)

        self.assertFalse(os.path.exists(self.output_file))
        expected_content = "Initial content"
        with open(self.input_file, 'r') as f:
            content = f.read()

        self.assertEqual(content.strip(), expected_content)

    def test_pipeline_with_quiet_failed_processor(self):
        class FailingProcessor:
            def process_file(self, source, destination):
                pass

        processors = [ProcessorA(), FailingProcessor(), ProcessorC()]
        pipeline = PipelineProcessor(True, processors=processors)
        pipeline.process_file(self.input_file, self.output_file)

        with open(self.output_file, 'r') as f:
            content = f.read()

        expected_content = "Initial content\nProcessed by A\nProcessed by C"
        self.assertEqual(content.strip(), expected_content)

    def test_pipeline_with_quiet_failed_processor_and_no_skipping(self):
        class FailingProcessor:
            def process_file(self, source, destination):
                pass

        processors = [ProcessorA(), FailingProcessor(), ProcessorC()]
        pipeline = PipelineProcessor(False, processors=processors)
        self.assertRaises(Exception, pipeline.process_file, self.input_file, self.output_file)

        self.assertFalse(os.path.exists(self.output_file))
        expected_content = "Initial content"
        with open(self.input_file, 'r') as f:
            content = f.read()

        self.assertEqual(content.strip(), expected_content)



class CompressionPostprocessor:
    __compressor_name: str

    def __init__(self, compressor_name: str):
        self.__compressor_name = compressor_name

    def postprocess(self, source_file: str, destination_file: str) -> None:
        print(f"** - Compressed Files with {self.__compressor_name}")



# package name PyMuPdf
import os

import fitz


def merge_pdf_files(file_list: list[str], merged_result_file: str) -> None:
    # merge page files into final destination
    merger = fitz.open()
    for file in file_list:
        if os.path.exists(merged_result_file):
            with fitz.open(merged_result_file) as f:
                merger.insert_pdf(f)
        with fitz.open(file) as f:
            merger.insert_pdf(f)
        os.remove(file)
    merger.save(merged_result_file)



import sys

from PIL.Image import DecompressionBombError

import os
# package name pillow
from PIL import Image
from img2pdf import convert

from ..config import TESSDATA_PREFIX, TESSERACT_PATH
from ..file_operations import get_filename
from ..processor import processor

# OCR for pdf
try:
    import pytesseract

    PY_TESS_AVAILABLE = True
except:
    PY_TESS_AVAILABLE = False


# TODO consoleUI processor


class ImagesToPdfConverter:
    def __init__(
            self,
            force_ocr: bool = False,
            no_ocr: bool = False,
            tesseract_language: str = "eng",
            event_handlers=None
    ):
        """
        :param force_ocr: (bool) force ocr even and fail if not possible
        :param no_ocr: (bool) skip ocr if not necessary or to preserve file size
        :param tesseract_language: (str) which language(s) are used for ocr
        :param event_handlers: (list of EventHandler) A list of event handlers. (pre and postprocessors) (pre and postprocessors)

        possible_files_endings = ['.jpc', '.xbm', '.j2k', '.icns', '.png', '.bmp', '.jp2', '.apng', '.j2c', '.jpe', '.tif',
                          '.rgb', '.tiff', '.jpx', '.bw', '.gif', '.jpg', '.jpf', '.jfif', '.rgba', '.sgi', '.webp',
                          '.dib', '.jpeg']

        """
        self.event_handlers = event_handlers
        if self.event_handlers is None:
            self.event_handlers = []

        if force_ocr and no_ocr:
            raise ValueError("force_ocr and no_ocr can't be used together")
        self.force_ocr = (force_ocr or not no_ocr) and PY_TESS_AVAILABLE
        self.no_ocr = no_ocr
        self.tesseract_language = tesseract_language

        try:
            self.__init_pytesseract(TESSERACT_PATH)
        except ValueError:
            self.force_ocr = False

    def __init_pytesseract(self, pytesseract_path) -> None:
        # either initiates pytesseract or deactivate ocr if not possible
        try:
            if not os.path.isfile(pytesseract_path):
                raise FileNotFoundError(r"[ ! ] - tesseract Path not found. Install "
                                        "https://github.com/UB-Mannheim/tesseract/wiki or edit "
                                        "'TESSERACT_PATH' to your specific tesseract executable")
            # set command (not sure why windows needs it differently)
            elif os.name == "nt":
                pytesseract.pytesseract.tesseract_cmd = f"{pytesseract_path}"
            else:
                pytesseract.tesseract_cmd = f"{pytesseract_path}"

        except Exception:
            if self.force_ocr:
                print("Tesseract Not Loaded, Can't create OCR."
                      "(leave out option '--ocr-force' to compress without ocr)", file=sys.stderr)
                self.force_ocr = False
            raise ValueError("Tesseract (-> no OCR on pdfs)")

    @processor
    def process_file(self, source_file: str, destination_path: str) -> None:
        if os.path.isdir(destination_path):
            destination_path = os.path.join(destination_path, get_filename(source_file) + ".pdf")

        try:
            if not self.force_ocr or self.no_ocr:
                raise ValueError("skipping tesseract")

            result = pytesseract.image_to_pdf_or_hocr(
                Image.open(source_file), lang=self.tesseract_language,
                extension="pdf",
                config=rf"{TESSDATA_PREFIX}"
            )
        except DecompressionBombError as e:
            raise e
        except ValueError:  # if ocr/tesseract fails or is skipped on purpose
            result = convert(source_file)
            print("No OCR applied.", file=sys.stderr)

        with open(destination_path, "wb") as f:
            f.write(result)



import os
import subprocess

from .config import PNGQUANT_PATH
from .file_operations import compare_and_use_better_option, check_if_valid_image
from .processor import processor


class PNGQuantCompressor:
    __FILE_SIZE_INCREASED_ERROR: int = 98
    __IMAGE_QUALITY_BELOW_LIMIT_ERROR: int = 99

    def __init__(
            self,
            speed: int = 1,
            min_quality: int = 80,
            max_quality: int = 100,
            event_handlers=None
    ):
        """
        Compresses PNG images using pngquant.
        :param speed: Speed of compression, where 0 is the slowest (best quality) and 10 is the fastest.
        :param min_quality: The minimum quality threshold, ranging from 1 to 99. If the conversion results
            in a quality below this threshold, the image won't be saved.
        :param max_quality: The maximum quality threshold, ranging from 1 to 99. Instructs pngquant to use
            the least amount of colors required to meet or exceed this quality.
        :param event_handlers: A list of event handlers. (pre and postprocessors)

        :returns: None
        """
        self.event_handlers = event_handlers
        if self.event_handlers is None:
            self.event_handlers = []

        pngquant_path = PNGQUANT_PATH
        if not os.path.isfile(pngquant_path):
            linux_error = "Install it with 'sudo apt install pngquant'." if os.name != "nt" else ""
            raise FileNotFoundError(
                rf"pngquant path not found at '{pngquant_path}'. {linux_error}")

        if speed < 0 or speed > 10:
            raise ValueError("speed needs to be a value in range 0-10")
        if min_quality < 0 or min_quality >= 100:
            raise ValueError("min_quality needs to be between 0 and 100")
        if max_quality < 0 or max_quality < min_quality:
            raise ValueError(
                "max_quality need to be greater than 0 and min_quality")

        system_extra = "powershell.exe" if os.name == 'nt' else ""
        pngquant_options = " ".join((
            f"--quality={min_quality}-{max_quality}",
            "--force",
            f"--speed {speed}",
            "--strip",
            "--ext '-comp.png'"
        ))
        self.pngquant_command = rf'{system_extra}  {pngquant_path} {pngquant_options}'

    @processor
    def process_file(self, source_file: str, destination_path: str):
        check_if_valid_image(source_file)
        if os.path.isdir(destination_path):
            destination_path = os.path.join(destination_path, os.path.basename(source_file))
        if not os.path.exists(os.path.dirname(destination_path)):
            os.makedirs(os.path.dirname(destination_path), exist_ok=True)

        subprocess.check_output(f'{self.pngquant_command} "{source_file}"',
                                stderr=subprocess.STDOUT, shell=True)
        result_file = source_file[:-4] + '-comp.png'
        compare_and_use_better_option(
            source_file, result_file, destination_path)
        if os.path.exists(result_file):
            os.remove(result_file)



import os
from unittest import TestCase
import file_crusher


class TestEventHandler:
    def __init__(self):
        self.preprocessor_calls = 0
        self.postprocessor_calls = 0


class TestEventHandlerPreProcessor(TestEventHandler):
    def preprocess(self, _, __):
        self.preprocessor_calls += 1


class TestEventHandlerPostProcessor(TestEventHandler):
    def postprocess(self, _, __):
        self.postprocessor_calls += 1


class TestEventHandlerPreAndPostProcessor(TestEventHandlerPreProcessor, TestEventHandlerPostProcessor):
    pass


class TestCPDFCompressor(TestCase):
    folder_url = os.path.join(os.path.dirname(__file__), "testdata.pdf")
    source_file_url = os.path.join(folder_url, "test_file.pdf")
    destination_file_url = os.path.join(folder_url, "test_file.pdf")

    def _check_and_delete_destination_file(self):
        self.assertTrue(os.path.isfile(self.destination_file_url))
        self.assertTrue(os.path.getsize(self.source_file_url) > os.path.getsize(self.destination_file_url))
        os.remove(self.destination_file_url)

    def test_extra_args_empty_string(self):
        self.assertFalse(os.path.isfile(self.destination_file_url))
        file_crusher.CPdfSqueezeCompressor(extra_args="").process_file(self.source_file_url, self.destination_file_url)
        self._check_and_delete_destination_file()

    def test_extra_args_some_arguments(self):
        self.assertFalse(os.path.isfile(self.destination_file_url))
        file_crusher.CPdfSqueezeCompressor(extra_args="-upw testpw").process_file(self.source_file_url,
                                                                                  self.destination_file_url)
        self._check_and_delete_destination_file()

    def test_extra_args_incorrect(self):
        with self.assertRaises(ValueError):
            file_crusher.CPdfSqueezeCompressor(extra_args="incorrect").process_file(self.source_file_url,
                                                                                    self.destination_file_url)

    def test_event_handlers_none(self):
        self.assertFalse(os.path.isfile(self.destination_file_url))
        file_crusher.CPdfSqueezeCompressor(event_handlers=None).process_file(self.source_file_url,
                                                                             self.destination_file_url)
        self._check_and_delete_destination_file()

    def test_event_handlers_one(self):
        self.assertFalse(os.path.isfile(self.destination_file_url))
        event_handler = TestEventHandlerPreAndPostProcessor()
        file_crusher.CPdfSqueezeCompressor(event_handlers=[event_handler]).process_file(self.source_file_url,
                                                                                        self.destination_file_url)
        self._check_and_delete_destination_file()
        self.assertTrue(event_handler.postprocessor_calls == 1)
        self.assertTrue(event_handler.preprocessor_calls == 1)

    def test_event_handlers_multiple(self):
        self.assertFalse(os.path.isfile(self.destination_file_url))
        event_handler1 = TestEventHandlerPreAndPostProcessor()
        event_handler2 = TestEventHandlerPreAndPostProcessor()
        file_crusher.CPdfSqueezeCompressor(event_handlers=[event_handler1, event_handler2]).process_file(self.source_file_url,
                                                                                        self.destination_file_url)
        self._check_and_delete_destination_file()
        self.assertTrue(event_handler1.postprocessor_calls == 1)
        self.assertTrue(event_handler1.preprocessor_calls == 1)
        self.assertTrue(event_handler2.postprocessor_calls == 1)
        self.assertTrue(event_handler2.preprocessor_calls == 1)

    def test_event_handlers_wrong_type(self):
        with self.assertRaises(TypeError):
            file_crusher.CPdfSqueezeCompressor(event_handlers="incorrect")

    def test_event_handlers_mix(self):
        with self.assertRaises(TypeError):
            event_handler = TestEventHandlerPreAndPostProcessor()
            file_crusher.CPdfSqueezeCompressor(event_handlers=[event_handler, "incorrect"])

    def test_process_file_file_to_file(self):
        compressor = file_crusher.CPdfSqueezeCompressor()
        source_file = self.source_file_url
        destination_path = os.path.join(self.folder_url, "compressed_file.pdf")
        compressor.process_file(source_file, destination_path)
        self.assertTrue(os.path.exists(destination_path))

    def test_process_file_file_to_folder(self):
        compressor = file_crusher.CPdfSqueezeCompressor()
        source_file = self.source_file_url
        destination_folder = self.folder_url
        compressor.process_file(source_file, destination_folder)
        compressed_file_path = os.path.join(destination_folder, os.path.basename(source_file))
        self.assertTrue(os.path.exists(compressed_file_path))

    def test_cpdf_use_wine_on_linux_is_false(self):
        compressor = file_crusher.PDFCompressor(compression_mode=1, default_pdf_dpi=200, force_ocr=True)
        source_file = "/home/user/Phone/tmp_BAFÖG/tmp2/Bafög 2023,2024.pdf"
        destination_file = "/home/user/Phone/tmp_BAFÖG/tmp2/Bafög 2023,2024_new.pdf"
        compressor.process_file(source_file, destination_file)
        self.assertTrue(os.path.exists(destination_file))

    def test_cpdf_use_wine_on_linux_is_true(self):
        raise NotImplementedError("TODO")

    def test_cpdf_extra_args_empty(self):
        raise NotImplementedError("TODO")

    def test_cpdf_extra_args_none(self):
        raise NotImplementedError("TODO")

    def test_cpdf_extra_args_working_command(self):
        raise NotImplementedError("TODO")

    def test_cpdf_extra_args_incorrect_command(self):
        raise NotImplementedError("TODO")

    def test_cpdf_event_handlers_none(self):
        raise NotImplementedError("TODO")

    def test_cpdf_event_handlers_one_call(self):
        raise NotImplementedError("TODO")

    def test_cpdf_event_handlers_multiple_calls(self):
        raise NotImplementedError("TODO")

    def test_cpdf_event_handlers_empty_list(self):
        raise NotImplementedError("TODO")

    def test_cpdf_event_handlers_none_of_the_methods_implemented(self):
        raise NotImplementedError("TODO")

    def test_cpdf_event_handlers_only_preprocessor_implemented(self):
        raise NotImplementedError("TODO")

    def test_cpdf_event_handlers_only_postprocessor_implemented(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_relative_source_path(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_absolute_source_path(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_source_path_none(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_source_path_empty(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_source_path_ends_with_some_text(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_source_path_wrong_file_type(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_source_path_only_file_ending(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_source_path_doesnt_exist(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_relative_path_to_output_file(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_absolute_path_to_output_file(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_relative_path_to_output_folder(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_absolute_path_to_output_folder(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_output_path_none(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_output_path_empty(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_output_path_has_no_file_ending(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_output_path_has_wrong_file_ending(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_output_path_is_only_file_ending(self):
        raise NotImplementedError("TODO")

    def test_cpdf_with_output_path_already_exists(self):
        raise NotImplementedError("TODO")

    def tearDown(self):
        # Clean up any files or folders created during testing
        pass

    # "../test_data/testFile.pdf"

    """
    Constructor
        use_wine_on_linux:
            True
            False
    
        extra_args
            ""
            None
            Something else like -upw
            incorrect one
    
        event_handlers
            None
            One
            Multiple
            []
            wrong type of class
            mix of wrong and correct classes
    
    process_file:
        source_file:
        destination_path:
            file -> file
            file -> folder
            
    """



import os
import shutil
import tempfile
from unittest import TestCase

from file_crusher import PNGQuantCompressor, PNGCompressor, PNGCrushCompressor, ADVPNGCompressor


class PngCompressorTest(TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.input_file = os.path.join(self.test_dir, 'input.png')
        self.output_file = os.path.join(self.test_dir, 'output.png')
        shutil.copyfile(os.path.join("test_data", "testimage.png"), self.input_file)

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_pngquant_processor_single_file(self):
        compressor = PNGQuantCompressor()
        self.assertFalse(os.path.exists(self.output_file))
        compressor.process_file(self.input_file, self.output_file)
        self.assertTrue(os.path.exists(self.output_file))

    def test_pngcrush_processor_single_file(self):
        compressor = PNGCrushCompressor()
        compressor.process_file(self.input_file, self.output_file)
        self.assertTrue(os.path.exists(self.output_file))

    def test_advpng_processor_single_file(self):
        compressor = ADVPNGCompressor()
        compressor.process_file(self.input_file, self.output_file)
        self.assertTrue(os.path.exists(self.output_file))

    def test_png_processor_single_file(self):
        compressor = PNGCompressor()
        compressor.process_file(self.input_file, self.output_file)
        self.assertTrue(os.path.exists(self.output_file))



from functools import wraps


def processor(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        for handler in self.event_handlers:
            if hasattr(handler, "preprocess"):
                handler.preprocess(*args, **kwargs)

        try:
            ret = method(self, *args, **kwargs)
        except Exception as e:
            raise RuntimeError(f"[!] Processing Failed during {self.__class__.__name__} stage.") from e

        for handler in self.event_handlers:
            if hasattr(handler, "postprocess"):
                handler.postprocess(*args, **kwargs)
        return ret

    return wrapper



import os
import shutil
import tempfile

from .processor import processor


class PipelineProcessor:
    def __init__(self, skip_if_no_result_file, processors=None, event_handlers=None):
        """
        Process files wih multiple processors in a row
        @param skip_if_no_result_file: if a processor fails and doesn't give a result file just skip it and continue with the next one
        @param processors: a list of processors
        """
        self._processors = processors
        if processors is None:
            self._processors = []
        self._skip_if_no_result_file = skip_if_no_result_file
        self.event_handlers = event_handlers
        if event_handlers is None:
            self.event_handlers = []

    @processor
    def process_file(self, source_file: str, destination_file: str, file_ending: str="auto") -> None:
        current_source = source_file
        temp_files = []
        if file_ending == "auto":
            file_ending = source_file.split(".")[-1]

        try:
            with tempfile.TemporaryDirectory() as tempdir:
                for i, processor in enumerate(self._processors):
                    temp_destination = os.path.join(tempdir, f"temp_{i}"+file_ending)
                    try:
                        processor.process_file(current_source, temp_destination)
                        if os.path.exists(temp_destination):
                            current_source = temp_destination
                            temp_files.append(temp_destination)
                        elif not self._skip_if_no_result_file:
                            raise Exception(f"Processor {processor.__class__.__name__} failed")

                    except Exception as e:
                        if self._skip_if_no_result_file:
                            print(f"Processor {processor.__class__.__name__} failed with error: {e}")
                            continue
                        else:
                            raise e
                shutil.copy(current_source, destination_file)
        finally:
            # Cleanup temporary files
            for temp_file in temp_files:
                if os.path.exists(temp_file):
                    os.remove(temp_file)
